@@ -59,6 +59,11 @@ from qgis.PyQt.QtWidgets import *
 from osgeo import gdal, ogr, osr, gdal_array
 import numpy as np
 from qgis.PyQt.QtWidgets import QAction, QMenu, QToolButton, QDialogButtonBox, QLabel, QGridLayout, QMainWindow
+
+try:
+    from .. import qps
+except:
+    import qps
 from . import DIR_UI_FILES
 
 # dictionary to store form classes and avoid multiple calls to read <myui>.i
@@ -66,11 +71,6 @@ QGIS_RESOURCE_WARNINGS = set()
 
 REMOVE_setShortcutVisibleInContextMenu = hasattr(QAction, 'setShortcutVisibleInContextMenu')
 
-try:
-    from .. import qps
-except:
-    import qps
-
 jp = os.path.join
 dn = os.path.dirname
 
@@ -130,6 +130,7 @@ def cleanDir(d):
         for p in dirs + files: rm(jp(root, p))
         break
 
+
 # a QPS internal map layer store
 QPS_MAPLAYER_STORE = QgsMapLayerStore()
 
@@ -455,7 +456,6 @@ LUT_WAVELENGTH = dict({'B': 480,
                        'SWIR2': 2150
                        })
 
-
 NEXT_COLOR_HUE_DELTA_CON = 10
 NEXT_COLOR_HUE_DELTA_CAT = 100
 
@@ -577,6 +577,8 @@ def createQgsField(name: str, exampleValue: typing.Any, comment: str = None) ->
         return QgsField(name, QVariant.String, 'varchar', comment=comment)
     elif isinstance(exampleValue, np.datetime64):
         return QgsField(name, QVariant.String, 'varchar', comment=comment)
+    elif isinstance(exampleValue, (bytes, QByteArray)):
+        return QgsField(name, QVariant.ByteArray, 'Binary', comment=comment)
     elif isinstance(exampleValue, list):
         assert len(exampleValue) > 0, 'need at least one value in provided list'
         v = exampleValue[0]
@@ -681,9 +683,11 @@ def gdalDataset(dataset: typing.Union[str,
                                       pathlib.Path,
                                       QgsRasterLayer,
                                       QgsRasterDataProvider,
-                                      gdal.Dataset], eAccess=gdal.GA_ReadOnly) -> gdal.Dataset:
+                                      gdal.Dataset],
+                eAccess: int = gdal.GA_ReadOnly) -> gdal.Dataset:
     """
     Returns a gdal.Dataset object instance
+    :param dataset:
     :param pathOrDataset: path | gdal.Dataset | QgsRasterLayer | QgsRasterDataProvider
     :return: gdal.Dataset
     """
@@ -696,9 +700,12 @@ def gdalDataset(dataset: typing.Union[str,
     if isinstance(dataset, QgsRasterDataProvider):
         return gdalDataset(dataset.dataSourceUri())
     if isinstance(dataset, str):
-        dataset = gdal.Open(dataset, eAccess)
+        ds = gdal.Open(dataset, eAccess)
+        assert isinstance(ds, gdal.Dataset), f'Can not read {dataset} as gdal.Dataset'
+        return ds
+    else:
+        raise NotImplementedError(f'Can not open {dataset} as gdal.Dataset')
 
-    assert isinstance(dataset, gdal.Dataset), 'Can not read {} as gdal.Dataset'.format(dataset)
     return dataset
 
 
@@ -896,6 +903,24 @@ def qgsRasterLayer(source) -> QgsRasterLayer:
     raise Exception('Unable to transform {} into QgsRasterLayer'.format(source))
 
 
+def qgsRasterLayers(sources) -> typing.Iterator[QgsRasterLayer]:
+    """
+    Like qgsRasterLayer, but on multiple inputs and with extraction of sub-layers
+    :param sources:
+    :return:
+    """
+    if not isinstance(sources, list):
+        sources = [sources]
+    assert isinstance(sources, list)
+
+    for source in sources:
+        lyr: QgsRasterLayer = qgsRasterLayer(source)
+        if lyr.isValid():
+            yield lyr
+        for lyr in qgsRasterLayers(lyr.subLayers()):
+            yield lyr
+
+
 def qgsMapLayer(value: typing.Any) -> QgsMapLayer:
     """
     Tries to convert the input into a QgsMapLayer
@@ -920,6 +945,7 @@ def qgsMapLayer(value: typing.Any) -> QgsMapLayer:
 
     return None
 
+
 def loadUi(uifile, baseinstance=None, package='', resource_suffix='_rc', remove_resource_references=True,
            loadUiType=False):
     """
@@ -1318,7 +1344,7 @@ def days_per_year(year):
     1. If the year is evenly divisible by 4, go to step 2. Otherwise, False.
     2. If the year is evenly divisible by 100, go to step 3. Otherwise, False
     3. If the year is evenly divisible by 400, True Otherwise, False
-    
+
     """
     """
     Every year that is exactly divisible by four is a leap year, except for years that are exactly divisible by 100, 
@@ -1392,9 +1418,9 @@ def defaultBands(dataset) -> list:
     elif isinstance(dataset, QgsRasterDataProvider):
         return defaultBands(dataset.dataSourceUri())
     elif isinstance(dataset, QgsRasterLayer) and \
-         isinstance(dataset.dataProvider(), QgsRasterDataProvider) and \
-         dataset.dataProvider().name() == 'gdal':
-            return defaultBands(dataset.source())
+            isinstance(dataset.dataProvider(), QgsRasterDataProvider) and \
+            dataset.dataProvider().name() == 'gdal':
+        return defaultBands(dataset.source())
     elif isinstance(dataset, gdal.Dataset):
 
         # check ENVI style metadata default band definition
@@ -1517,7 +1543,7 @@ def parseFWHM(dataset) -> typing.Tuple[np.ndarray]:
         # search band by band
         values = []
         for b in range(dataset.RasterCount):
-            band: gdal.Band = dataset.GetRasterBand(b+1)
+            band: gdal.Band = dataset.GetRasterBand(b + 1)
             for key, domain in key_positions:
                 value = dataset.GetMetadataItem(key, domain)
                 if value not in ['', None]:
@@ -1527,6 +1553,55 @@ def parseFWHM(dataset) -> typing.Tuple[np.ndarray]:
             return np.asarray(values)
     return None
 
+
+def checkWavelength(key: str, values: str, expected: int = 1) -> np.ndarray:
+    wl: np.ndarray = None
+    if re.search(r'^(center[ _]*)?wavelengths?$', key, re.I):
+        # remove trailing / ending { } and whitespace
+        values = re.sub('[{}]', '', values).strip()
+        if ',' not in values:
+            sep = ' '
+        else:
+            sep = ','
+        try:
+            wl = np.fromstring(values, count=expected, sep=sep)
+        except ValueError as exV:
+            pass
+        except Exception as ex:
+            pass
+    return wl
+
+
+def checkWavelengthUnit(key: str, value: str) -> str:
+    wlu: str = None
+    value = value.strip()
+    if re.search(r'^wavelength[ _]?units?', key, re.I):
+        # metric length units
+        wlu = UnitLookup.baseUnit(value)
+
+        if wlu is not None:
+            return wlu
+
+        if re.search(r'^Wavenumber$', value, re.I):
+            wlu = '-'
+        elif re.search(r'^GHz$', value, re.I):
+            wlu = 'GHz'
+        elif re.search(r'^MHz$', value, re.I):
+            wlu = 'MHz'
+        # date / time units
+        elif re.search(r'^(Date|DTG|Date[_ ]?Time[_ ]?Group|Date[_ ]?Stamp|Time[_ ]?Stamp)$', value, re.I):
+            wlu = 'DateTime'
+        elif re.search(r'^Decimal[_ ]?Years?$', value, re.I):
+            wlu = 'DecimalYear'
+        elif re.search(r'^(Seconds?|s|secs?)$', value, re.I):
+            wlu = 's'
+        elif re.search(r'^Index$', value, re.I):
+            wlu = None
+        else:
+            wlu = None
+    return wlu
+
+
 def parseWavelength(dataset) -> typing.Tuple[np.ndarray, str]:
     """
     Returns the wavelength + wavelength unit of a raster
@@ -1534,120 +1609,99 @@ def parseWavelength(dataset) -> typing.Tuple[np.ndarray, str]:
     :return: (wl, wl_u) or (None, None), if not existing
     """
 
-    wl = None
-    wlu = None
+    def sort_domains(domains) -> typing.List[str]:
+        if not isinstance(domains, list):
+            domains = []
+        return sorted(domains, key=lambda n: n != ['ENVI'])
+
     try:
         dataset = gdalDataset(dataset)
-    except:
-        pass
-
-    def checkWavelengthUnit(key: str, value: str) -> str:
-        wlu = None
-        value = value.strip()
-        if re.search(r'wavelength.units?', key, re.I):
-            # metric length units
-            if re.search(r'^(Micrometers?|um|μm)$', values, re.I):
-                wlu = 'μm'  # fix with python 3 UTF
-            elif re.search(r'^(Nanometers?|nm)$', values, re.I):
-                wlu = 'nm'
-            elif re.search(r'^(Millimeters?|mm)$', values, re.I):
-                wlu = 'nm'
-            elif re.search(r'^(Centimeters?|cm)$', values, re.I):
-                wlu = 'nm'
-            elif re.search(r'^(Meters?|m)$', values, re.I):
-                wlu = 'nm'
-            elif re.search(r'^Wavenumber$', values, re.I):
-                wlu = '-'
-            elif re.search(r'^GHz$', values, re.I):
-                wlu = 'GHz'
-            elif re.search(r'^MHz$', values, re.I):
-                wlu = 'MHz'
-            # date / time units
-            elif re.search(r'^(Date|DTG|Date[_ ]?Time[_ ]?Group|Date[_ ]?Stamp|Time[_ ]?Stamp)$', values, re.I):
-                wlu = 'DateTime'
-            elif re.search(r'^Decimal[_ ]?Years?$', value, re.I):
-                wlu = 'DecimalYear'
-            elif re.search(r'^(Seconds?|s|secs?)$', values, re.I):
-                wlu = 's'
-            elif re.search(r'^Index$', values, re.I):
-                wlu = None
-            else:
-                wlu = None
-        return wlu
-
-    def checkWavelength(key: str, values: str):
-        wl = None
-        if re.search(r'wavelengths?$', key, re.I):
-            # remove trailing / ending { } and whitespace
-            values = re.sub('[{}]', '', values).strip()
-            if ',' not in values:
-                sep = ' '
-            else:
-                sep = ','
-            try:
-                wl = np.fromstring(values, count=dataset.RasterCount, sep=sep)
-            except ValueError as exV:
-                pass
-            except Exception as ex:
-                pass
-        return wl
+    except AssertionError:
+        return None, None
 
     if isinstance(dataset, gdal.Dataset):
         # 1. check on raster level
-        domains = dataset.GetMetadataDomainList()
-        if isinstance(domains, list):
-            for domain in domains:
-                # see http://www.harrisgeospatial.com/docs/ENVIHeaderFiles.html for supported wavelength units
+        for domain in sort_domains(dataset.GetMetadataDomainList()):
+            # see http://www.harrisgeospatial.com/docs/ENVIHeaderFiles.html for supported wavelength units
 
-                mdDict = dataset.GetMetadata_Dict(domain)
+            mdDict = dataset.GetMetadata_Dict(domain)
 
-                for key, values in mdDict.items():
-                    if wl is None:
-                        wl = checkWavelength(key, values)
-                    if wlu is None:
-                        wlu = checkWavelengthUnit(key, values)
-                    if wl is not None and wlu is not None:
-                        # domain-specific check
-                        if domain == 'FORCE' and wlu == 'DecimalYear':
-                            # make decimal-year values leap-year sensitive
-                            wl = convertDateUnit(datetime64(wl, dpy=365), 'DecimalYear')
-                        break
+            domainWLU: str = None
+            domainWL: np.ndarray = None
+            # search domain
+            for key, values in mdDict.items():
+                if domainWL is None:
+                    domainWL = checkWavelength(key, values, expected=dataset.RasterCount)
+                if domainWLU is None:
+                    domainWLU = checkWavelengthUnit(key, values)
 
-        if wl is not None and wlu is not None:
-            if len(wl) > dataset.RasterCount:
-                wl = wl[0:dataset.RasterCount]
-            return wl, wlu
+            if isinstance(domainWL, np.ndarray) and isinstance(domainWLU, str):
+                if domain == 'FORCE' and domainWLU == 'DecimalYear':
+                    # make decimal-year values leap-year sensitive
+                    domainWL = convertDateUnit(datetime64(domainWL, dpy=365), 'DecimalYear')
+
+                if len(domainWL) > dataset.RasterCount:
+                    domainWL = domainWL[0:dataset.RasterCount]
+
+                return domainWL, domainWLU
+
+        # 2. check on band level. collect wl from each single band
+        # first domain that defines wl and wlu is prototyp domain for all other bands
+
+        wl = []  # list of wavelength values
+        wlu: str = None  # wavelength unit string
+        wlDomain: str = None  # the domain in which the WL and WLU are defined
+        wlKey: str = None  # key that stores the wavelength value
+        wluKey: str = None  # key that stores the wavelength unit
 
-        # 2. check on band level
-        wl = []
         for b in range(dataset.RasterCount):
             band: gdal.Band = dataset.GetRasterBand(b + 1)
-            domains = band.GetMetadataDomainList()
-            if isinstance(domains, list):
-                for domain in domains:
+            if b == 0:
+                for domain in sort_domains(band.GetMetadataDomainList()):
                     # see http://www.harrisgeospatial.com/docs/ENVIHeaderFiles.html for supported wavelength units
+                    domainWL = domainWLU = None
+
                     mdDict = band.GetMetadata_Dict(domain)
-                    _wl = None
+
                     for key, values in mdDict.items():
-                        if wlu is None:
-                            wlu = checkWavelengthUnit(key, values)
-                        if _wl is None:
-                            _wl = checkWavelength(key, values)
-                    if wlu is not None:
-                        s = ""
-                    if wlu is not None and _wl is not None:
-                        wl.append(_wl[0])
-                        break
+                        if domainWLU is None:
+                            domainWLU = checkWavelengthUnit(key, values)
+                            if domainWLU:
+                                wluKey = key
+
+                        if domainWL is None:
+                            domainWL = checkWavelength(key, values, expected=1)
+                            if isinstance(domainWL, np.ndarray):
+                                wlKey = key
+
+                    if isinstance(domainWL, np.ndarray) and isinstance(domainWLU, str):
+                        wlDomain = domain
+                        wlu = domainWLU
+                        wl.append(domainWL[0])
+
+                if len(wl) == 0:
+                    # we did not found a WL + WLU for the 1st band. stop searching and return
+                    return None, None
+            else:
+                bandWLU = checkWavelengthUnit(wluKey, band.GetMetadataItem(wluKey, wlDomain))
+                bandWL = checkWavelength(wlKey, band.GetMetadataItem(wlKey, wlDomain), expected=1)
+
+                if bandWLU != wlu or bandWL is None:
+                    print(f'{dataset.GetDescription()}: inconsistent use of metadata key {wluKey} per band')
+                    return None, None
+                wl.append(bandWL[0])
 
         if len(wl) == 0:
-            wl = None
-        else:
-            wl = np.asarray(wl)
-            if domain == 'FORCE' and wlu == 'DecimalYear':
-                # make decimal-year values leap-year sensitive
-                wl = UnitLookup.convertDateUnit(datetime64(wl, dpy=365), 'DecimalYear')
+            return None, None
 
-    return wl, wlu
+        wl = np.asarray(wl)
+        if domain == 'FORCE' and wlu == 'DecimalYear':
+            # make decimal-year values leap-year sensitive
+            wl = UnitLookup.convertDateUnit(datetime64(wl, dpy=365), 'DecimalYear')
+
+        return wl, wlu
+    else:
+        return None, None
 
 
 class Singleton(type):
@@ -1894,6 +1948,7 @@ class SpatialPoint(QgsPointXY):
     """
     Object to keep QgsPoint and QgsCoordinateReferenceSystem together
     """
+
     @staticmethod
     def readXml(node: QDomNode):
         wkt = node.firstChildElement('SpatialPointCrs').text()
@@ -2086,6 +2141,7 @@ class SpatialExtent(QgsRectangle):
     """
     Object that combines a QgsRectangle and QgsCoordinateReferenceSystem
     """
+
     @staticmethod
     def readXml(node: QDomNode):
         wkt = node.firstChildElement('SpatialExtentCrs').text()
@@ -2094,7 +2150,7 @@ class SpatialExtent(QgsRectangle):
         return SpatialExtent(crs, rectangle)
 
     @staticmethod
-    def fromMapCanvas(mapCanvas, fullExtent:bool=False):
+    def fromMapCanvas(mapCanvas, fullExtent: bool = False):
         assert isinstance(mapCanvas, QgsMapCanvas)
 
         if fullExtent:
@@ -2340,6 +2396,7 @@ def setToolButtonDefaultActionMenu(toolButton: QToolButton, actions: list):
     menu.triggered.connect(toolButton.setDefaultAction)
     toolButton.setMenu(menu)
 
+
 class SelectMapLayersDialog(QgsDialog):
     class LayerDescription(object):
 
